# -*- coding: utf-8 -*-

"""
SQL statements are kept here as to not clutter up main file.
"""


VACUUM_STATS = """
    WITH table_opts AS (
        SELECT
            pg_class.oid,
            relname,
            nspname,
            array_to_string(reloptions, '') AS relopts
        FROM pg_class
            INNER JOIN pg_namespace ns ON relnamespace = ns.oid
    ), vacuum_settings AS (
        SELECT
            oid,
            relname,
            nspname,
            CASE
                WHEN relopts LIKE '%autovacuum_vacuum_threshold%'
                THEN regexp_replace(
                    relopts,
                    '.*autovacuum_vacuum_threshold=([0-9.]+).*',
                    E'\\\\\\1'
                )::integer
                ELSE current_setting('autovacuum_vacuum_threshold')::integer
            END AS autovacuum_vacuum_threshold,
            CASE
                WHEN relopts LIKE '%autovacuum_vacuum_scale_factor%'
                THEN regexp_replace(
                    relopts,
                    '.*autovacuum_vacuum_scale_factor=([0-9.]+).*',
                    E'\\\\\\1'
                )::real
                ELSE current_setting('autovacuum_vacuum_scale_factor')::real
            END AS autovacuum_vacuum_scale_factor
        FROM table_opts
    )
    SELECT
        vacuum_settings.nspname AS schema,
        vacuum_settings.relname AS table,
        to_char(psut.last_vacuum, 'YYYY-MM-DD HH24:MI') AS last_vacuum,
        to_char(psut.last_autovacuum, 'YYYY-MM-DD HH24:MI') AS last_autovacuum,
        to_char(pg_class.reltuples, '9G999G999G999') AS rowcount,
        to_char(psut.n_dead_tup, '9G999G999G999') AS dead_rowcount,
        to_char(
            autovacuum_vacuum_threshold + (
                autovacuum_vacuum_scale_factor::numeric * pg_class.reltuples),
                '9G999G999G999'
        ) AS autovacuum_threshold,
        CASE
            WHEN autovacuum_vacuum_threshold + (
                autovacuum_vacuum_scale_factor::numeric * pg_class.reltuples
            ) < psut.n_dead_tup
            THEN 'yes'
        END AS expect_autovacuum
    FROM pg_stat_user_tables psut
        INNER JOIN pg_class ON psut.relid = pg_class.oid
        INNER JOIN vacuum_settings ON pg_class.oid = vacuum_settings.oid
    ORDER BY 1
"""

OUTLIERS = """
    SELECT {query} AS qry,
        interval '1 millisecond' * total_time AS exec_time,
        to_char((total_time/sum(total_time) OVER()) * 100,
            'FM90D0') || '%' AS
        prop_exec_time,
        to_char(calls, 'FM999G999G990') AS ncalls,
        interval '1 millisecond' * (blk_read_time + blk_write_time)
            AS sync_io_time
    FROM pg_stat_statements
    WHERE userid = (
        SELECT usesysid
        FROM pg_user
        WHERE usename = current_user
        LIMIT 1
    )
    ORDER BY total_time DESC
    LIMIT 10
"""

BLOCKING = """
    SELECT
        bl.pid AS blocked_pid,
        ka.{query_column} AS blocking_statement,
        now() - ka.query_start AS blocking_duration,
        kl.pid AS blocking_pid,
        a.{query_column} AS blocked_statement,
        now() - a.query_start AS blocked_duration
    FROM
        pg_catalog.pg_locks bl
        JOIN pg_catalog.pg_stat_activity a ON bl.pid = a.{pid_column}
        JOIN pg_catalog.pg_locks kl
        JOIN pg_catalog.pg_stat_activity ka ON kl.pid = ka.{pid_column}
            ON bl.transactionid = kl.transactionid AND bl.pid != kl.pid
    WHERE NOT bl.granted
"""

INDEX_USAGE = """
    SELECT
        relname,
        CASE idx_scan
            WHEN 0 THEN 'Insufficient data'
            ELSE (100 * idx_scan / (seq_scan + idx_scan))::text
        END percent_of_times_index_used,
        n_live_tup rows_in_table
    FROM pg_stat_user_tables
    ORDER BY n_live_tup DESC
"""

CALLS = """
    {select} interval '1 millisecond' * total_time AS exec_time,
        to_char((total_time/sum(total_time) OVER()) * 100, 'FM90D0') || '%'
            AS prop_exec_time,
        to_char(calls, 'FM999G999G990') AS ncalls,
        interval '1 millisecond' * (blk_read_time + blk_write_time)
            AS sync_io_time
    FROM pg_stat_statements
    WHERE userid = (
        SELECT usesysid
        FROM pg_user
        WHERE usename = current_user LIMIT 1
    )
    ORDER BY calls DESC
    LIMIT 10
"""

LOCKS = """
    SELECT
        pg_stat_activity.{pid_column},
        pg_class.relname,
        pg_locks.transactionid,
        pg_locks.granted,
        substr(pg_stat_activity.{query_column},1,30) AS query_snippet,
        age(now(),pg_stat_activity.query_start) AS "age"
    FROM
        pg_stat_activity,
        pg_locks LEFT OUTER JOIN pg_class ON (pg_locks.relation = pg_class.oid)
    WHERE
        pg_stat_activity.{query_column} <> '<insufficient privilege>'
            AND pg_locks.pid = pg_stat_activity.{pid_column}
            AND pg_locks.mode = 'ExclusiveLock'
            AND pg_stat_activity.{pid_column} <> pg_backend_pid()
    ORDER BY query_start
"""

PG_STAT_STATEMENT = """
    SELECT exists(
        SELECT 1
        FROM pg_extension
        WHERE extname = 'pg_stat_statements'
    ) AS available
"""

BLOAT = """
    WITH constants AS (
        SELECT
            current_setting('block_size')::numeric AS bs,
            23 AS hdr,
            4 AS ma
    ), bloat_info AS (
        SELECT
            ma,
            bs,
            schemaname,
            tablename,
            (datawidth+(
                hdr+ma-(
                    CASE
                        WHEN hdr%ma=0
                        THEN ma
                        ELSE hdr%ma
                    END
                )
            ))::numeric AS datahdr,
            (maxfracsum*(
                nullhdr+ma-(
                    CASE
                        WHEN nullhdr%ma=0
                        THEN ma
                        ELSE nullhdr%ma
                    END
                )
            )) AS nullhdr2
        FROM (
            SELECT
                schemaname,
                tablename,
                hdr,
                ma,
                bs,
                SUM((1-null_frac)*avg_width) AS datawidth,
                MAX(null_frac) AS maxfracsum,
                hdr+(
                    SELECT 1+count(*)/8
                    FROM pg_stats s2
                    WHERE
                        null_frac<>0
                        AND s2.schemaname = s.schemaname
                        AND s2.tablename = s.tablename
                ) AS nullhdr
            FROM
                pg_stats s,
                constants
            GROUP BY 1,2,3,4,5
        ) AS foo
    ), table_bloat AS (
        SELECT
            schemaname,
            tablename,
            cc.relpages,
            bs,
            CEIL(
                (cc.reltuples*(
                    (datahdr+ma-(
                        CASE
                            WHEN datahdr%ma=0
                            THEN ma
                            ELSE datahdr%ma
                        END
                    ))+nullhdr2+4
                ))/(bs-20::float)
            ) AS otta
        FROM bloat_info
            JOIN pg_class cc ON cc.relname = bloat_info.tablename
            JOIN pg_namespace nn ON cc.relnamespace = nn.oid
                AND nn.nspname = bloat_info.schemaname
                AND nn.nspname <> 'information_schema'
    ), index_bloat AS (
        SELECT
            schemaname,
            tablename,
            bs,
            COALESCE(c2.relname,'?') AS iname,
            COALESCE(c2.reltuples,0) AS ituples,
            COALESCE(c2.relpages,0) AS ipages,
            COALESCE(
                CEIL((c2.reltuples*(datahdr-12))/(bs-20::float)),0
            ) AS iotta
        FROM bloat_info
            JOIN pg_class cc ON cc.relname = bloat_info.tablename
            JOIN pg_namespace nn ON cc.relnamespace = nn.oid
                AND nn.nspname = bloat_info.schemaname
                AND nn.nspname <> 'information_schema'
            JOIN pg_index i ON indrelid = cc.oid
            JOIN pg_class c2 ON c2.oid = i.indexrelid
     )
     SELECT
        type,
        schemaname,
        object_name,
        bloat,
        pg_size_pretty(raw_waste) AS waste
     FROM (
        SELECT
            'table' AS type,
            schemaname,
            tablename AS object_name,
            ROUND(
                CASE
                    WHEN otta=0
                    THEN 0.0
                    ELSE table_bloat.relpages/otta::numeric
                END,
            1) AS bloat,
            CASE
                WHEN relpages < otta
                THEN '0'
                ELSE (bs*(
                    table_bloat.relpages-otta)::bigint)::bigint
            END AS raw_waste
        FROM table_bloat
        UNION
        SELECT
            'index' AS type,
            schemaname,
            tablename || '::' || iname AS object_name,
            ROUND(
                CASE
                    WHEN iotta=0 OR ipages=0
                    THEN 0.0
                    ELSE ipages/iotta::numeric
                END,
            1) AS bloat,
            CASE
                WHEN ipages < iotta
                THEN '0'
                ELSE (bs*(ipages-iotta))::bigint
            END AS raw_waste
        FROM index_bloat
    ) bloat_summary
    ORDER BY
        raw_waste DESC,
        bloat DESC
"""

LONG_RUNNING_QUERIES = """
     SELECT
        {pid_column},
        now() - pg_stat_activity.query_start AS duration,
        {query_column} AS query
    FROM pg_stat_activity
    WHERE
        pg_stat_activity.{query_column} <> ''::text
        {idle}
        AND now() - pg_stat_activity.query_start > interval '5 minutes'
    ORDER BY now() - pg_stat_activity.query_start DESC
"""

SEQ_SCANS = """
     SELECT
        relname AS name,
        seq_scan AS count
     FROM pg_stat_user_tables
     ORDER BY seq_scan DESC
"""

UNUSED_INDEXES = """
    SELECT
        schemaname || '.' || relname AS table,
        indexrelname AS index,
        pg_size_pretty(pg_relation_size(i.indexrelid)) AS index_size,
        idx_scan AS index_scans
    FROM pg_stat_user_indexes ui
        JOIN pg_index i ON ui.indexrelid = i.indexrelid
    WHERE
        NOT indisunique AND idx_scan < 50
        AND pg_relation_size(relid) > 5 * 8192
    ORDER BY
        pg_relation_size(i.indexrelid) / nullif(idx_scan, 0) DESC
        NULLS FIRST,
        pg_relation_size(i.indexrelid) DESC
"""

TOTAL_TABLE_SIZE = """
    SELECT
        c.relname AS name,
        pg_size_pretty(pg_total_relation_size(c.oid)) AS size
    FROM pg_class c
        LEFT JOIN pg_namespace n ON (n.oid = c.relnamespace)
    WHERE
        n.nspname NOT IN ('pg_catalog', 'information_schema')
        AND n.nspname !~ '^pg_toast'
        AND c.relkind='r'
    ORDER BY pg_total_relation_size(c.oid) DESC
"""

TOTAL_INDEXES_SIZE = """
    SELECT
        c.relname AS table,
        pg_size_pretty(pg_indexes_size(c.oid)) AS index_size
    FROM pg_class c
        LEFT JOIN pg_namespace n ON (n.oid = c.relnamespace)
    WHERE
        n.nspname NOT IN ('pg_catalog', 'information_schema')
        AND n.nspname !~ '^pg_toast'
        AND c.relkind='r'
    ORDER BY pg_indexes_size(c.oid) DESC;
"""

TABLE_SIZE = """
     SELECT
        c.relname AS name,
        pg_size_pretty(pg_table_size(c.oid)) AS size
     FROM pg_class c
        LEFT JOIN pg_namespace n ON (n.oid = c.relnamespace)
     WHERE
        n.nspname NOT IN ('pg_catalog', 'information_schema')
        AND n.nspname !~ '^pg_toast'
        AND c.relkind='r'
     ORDER BY pg_table_size(c.oid) DESC
"""

INDEX_SIZE = """
    SELECT
        c.relname AS name,
        pg_size_pretty(sum(c.relpages::bigint*8192)::bigint) AS size
    FROM pg_class c
        LEFT JOIN pg_namespace n ON (n.oid = c.relnamespace)
    WHERE
        n.nspname NOT IN ('pg_catalog', 'information_schema')
        AND n.nspname !~ '^pg_toast'
        AND c.relkind='i'
    GROUP BY c.relname
    ORDER BY sum(c.relpages) DESC
"""

TOTAL_INDEX_SIZE = """
    SELECT pg_size_pretty(sum(c.relpages::bigint*8192)::bigint) AS size
    FROM pg_class c
        LEFT JOIN pg_namespace n ON (n.oid = c.relnamespace)
    WHERE
        n.nspname NOT IN ('pg_catalog', 'information_schema')
        AND n.nspname !~ '^pg_toast'
        AND c.relkind='i';
"""

CACHE_HIT = """
    SELECT
        'index hit rate' AS name,
        (sum(idx_blks_hit)) / sum(idx_blks_hit + idx_blks_read) AS ratio
    FROM pg_statio_user_indexes
    UNION ALL
    SELECT
        'table hit rate' AS name,
        sum(heap_blks_hit) / (sum(heap_blks_hit) + sum(heap_blks_read))
            AS ratio
    FROM pg_statio_user_tables
"""

TABLE_INDEXES_SIZE = """
    SELECT
        c.relname AS table,
        pg_size_pretty(pg_indexes_size(c.oid)) AS index_size
    FROM pg_class c
        LEFT JOIN pg_namespace n ON (n.oid = c.relnamespace)
    WHERE n.nspname NOT IN ('pg_catalog', 'information_schema')
        AND n.nspname !~ '^pg_toast'
        AND c.relkind='r'
    ORDER BY pg_indexes_size(c.oid) DESC;
"""

PS = """
    SELECT
        {pid_column},
        application_name AS source,
        age(now(),query_start) AS running_for,
        waiting,
        {query_column} AS query
    FROM pg_stat_activity
    WHERE {query_column} <> '<insufficient privilege>'
        AND {pid_column} <> pg_backend_pid()
        {idle}
    ORDER BY query_start DESC
"""

VERSION = """
    SELECT version()
"""



# -*- coding: utf-8 -*-
import re

import psycopg2
import psycopg2.extras

from . import sql_constants as sql

__author__ = 'Scott Woodall'
__email__ = 'scott.woodall@gmail.com'
__version__ = '0.1.1'


class PgExtras(object):
    def __init__(self, dsn=None):
        self.dsn = dsn
        self._pg_stat_statement = None
        self._cursor = None
        self._conn = None
        self._is_pg_at_least_nine_two = None

    def __enter__(self):
        """
        The context manager convention is preferred so that if there are ever
        any exceptions the database cursor/connection will always be closed.
        """
        return self

    def __exit__(self, type, value, trace):
        self.close_db_connection()

    @property
    def cursor(self):
        if self._cursor is None:
            self._conn = psycopg2.connect(
                self.dsn,
                cursor_factory=psycopg2.extras.NamedTupleCursor
            )

            self._cursor = self._conn.cursor()

        return self._cursor

    @property
    def query_column(self):
        """
        PG9.2 changed column names.

        :returns: str
        """

        if self.is_pg_at_least_nine_two():
            return 'query'
        else:
            return 'current_query'

    @property
    def pid_column(self):
        """
        PG9.2 changed column names.

        :returns: str
        """

        if self.is_pg_at_least_nine_two():
            return 'pid'
        else:
            return 'procpid'

    def pg_stat_statement(self):
        """
        Some queries require the pg_stat_statement module to be installed.
        http://www.postgresql.org/docs/current/static/pgstatstatements.html

        :returns: boolean
        """

        if self._pg_stat_statement is None:
            results = self.execute(sql.PG_STAT_STATEMENT)
            is_available = results[0].available

            if is_available:
                self._pg_stat_statement = True
            else:
                raise Exception("""
                    pg_stat_statements extension needs to be installed in the
                    public schema first. This extension is only available on
                    Postgres versions 9.2 or greater. You can install it by
                    adding pg_stat_statements to shared_preload_libraries in
                    postgresql.conf, restarting postgres and then running the
                    following sql statement in your database:
                    CREATE EXTENSION pg_stat_statements;
                """)

        return self._pg_stat_statement

    def is_pg_at_least_nine_two(self):
        """
        Some queries have different syntax depending what version of postgres
        we are querying against.

        :returns: boolean
        """

        if self._is_pg_at_least_nine_two is None:
            results = self.version()
            regex = re.compile("PostgreSQL (\d+\.\d+\.\d+) on")
            matches = regex.match(results[0].version)
            version = matches.groups()[0]

            if version > '9.2.0':
                self._is_pg_at_least_nine_two = True
            else:
                self._is_pg_at_least_nine_two = False

        return self._is_pg_at_least_nine_two

    def close_db_connection(self):
        if self._cursor is not None:
            self._cursor.close()

        if self._conn is not None:
            self._conn.close()

    def execute(self, statement):
        """
        Execute the given sql statement.

        :param statement: sql statement to run
        :returns: list
        """

        # Make the sql statement easier to read in case some of the queries we
        # run end up in the output
        sql = statement.replace('\n', '')
        sql = ' '.join(sql.split())
        self.cursor.execute(sql)

        return self.cursor.fetchall()

    def cache_hit(self):
        """
        Calculates your cache hit rate (effective databases are at 99% and up).

        Record(
            name='index hit rate',
            ratio=Decimal('0.99994503346970922117')
        )

        :returns: list of Records
        """

        return self.execute(sql.CACHE_HIT)

    def index_usage(self):
        """
        Calculates your index hit rate (effective databases are at 99% and up).

        Record(
            relname='pgbench_history',
            percent_of_times_index_used=None,
            rows_in_table=249976
        )

        :returns: list of Records
        """

        return self.execute(sql.INDEX_USAGE)

    def calls(self, truncate=False):
        """
        Show 10 most frequently called queries. Requires the pg_stat_statements
        Postgres module to be installed.

        Record(
            query='BEGIN;',
            exec_time=datetime.timedelta(0, 0, 288174),
            prop_exec_time='0.0%',
            ncalls='845,590',
            sync_io_time=datetime.timedelta(0)
        )

        :param truncate: trim the Record.query output if greater than 40 chars
        :returns: list of Records
        """

        if self.pg_stat_statement():
            if truncate:
                select = """
                    SELECT CASE
                        WHEN length(query) < 40
                        THEN query
                        ELSE substr(query, 0, 38) || '..'
                    END AS qry,
                """
            else:
                select = 'SELECT query,'

            return self.execute(sql.CALLS.format(select=select))

    def blocking(self):
        """
        Display queries holding locks other queries are waiting to be
        released.

        Record(
            pid=40821,
            source='',
            running_for=datetime.timedelta(0, 0, 2857),
            waiting=False,
            query='SELECT pg_sleep(10);'
        )

        :returns: list of Records
        """

        return self.execute(
            sql.BLOCKING.format(
                query_column=self.query_column,
                pid_column=self.pid_column
            )
        )

    def outliers(self, truncate=False):
        """
        Show 10 queries that have longest execution time in aggregate. Requires
        the pg_stat_statments Postgres module to be installed.

        Record(
            qry='UPDATE pgbench_tellers SET tbalance = tbalance + ?;',
            exec_time=datetime.timedelta(0, 19944, 993099),
            prop_exec_time='67.1%',
            ncalls='845,589',
            sync_io_time=datetime.timedelta(0)
        )

        :param truncate: trim the Record.qry output if greater than 40 chars
        :returns: list of Records
        """

        if self.pg_stat_statement():
            if truncate:
                query = """
                    CASE WHEN length(query) < 40
                        THEN query
                        ELSE substr(query, 0, 38) || '..'
                    END
                """
            else:
                query = 'query'

            return self.execute(sql.OUTLIERS.format(query=query))

    def vacuum_stats(self):
        """
        Show dead rows and whether an automatic vacuum is expected to be
        triggered.

        Record(
            schema='public',
            table='pgbench_tellers',
            last_vacuum='2014-04-29 14:45',
            last_autovacuum='2014-04-29 14:45',
            rowcount='10',
            dead_rowcount='0',
            autovacuum_threshold='52',
            expect_autovacuum=None
        )

        :returns: list of Records
        """

        return self.execute(sql.VACUUM_STATS)

    def bloat(self):
        """
        Table and index bloat in your database ordered by most wasteful.

        Record(
            type='index',
            schemaname='public',
            object_name='pgbench_accounts::pgbench_accounts_pkey',
            bloat=Decimal('0.2'),
            waste='0 bytes'
        )

        :returns: list of Records
        """

        return self.execute(sql.BLOAT)

    def long_running_queries(self):
        """
        Show all queries longer than five minutes by descending duration.

        Record(
            pid=19578,
            duration=datetime.timedelta(0, 19944, 993099),
            query='SELECT * FROM students'
        )

        :returns: list of Records
        """

        if self.is_pg_at_least_nine_two():
            idle = "AND state <> 'idle'"
        else:
            idle = "AND current_query <> '<IDLE>'"

        return self.execute(
            sql.LONG_RUNNING_QUERIES.format(
                pid_column=self.pid_column,
                query_column=self.query_column,
                idle=idle
            )
        )

    def seq_scans(self):
        """
        Show the count of sequential scans by table descending by order.

        Record(
            name='pgbench_branches',
            count=237
        )

        :returns: list of Records
        """

        return self.execute(sql.SEQ_SCANS)

    def unused_indexes(self):
        """
        Show unused and almost unused indexes, ordered by their size relative
        to the number of index scans. Exclude indexes of very small tables
        (less than 5 pages), where the planner will almost invariably select
        a sequential scan, but may not in the future as the table grows.

        Record(
            table='public.grade_levels',
            index='index_placement_attempts_on_grade_level_id',
            index_size='97 MB',
            index_scans=0
        )

        :returns: list of Records
        """

        return self.execute(sql.UNUSED_INDEXES)

    def total_table_size(self):
        """
        Show the size of the tables (including indexes), descending by size.

        Record(
            name='pgbench_accounts',
            size='15 MB'
        )

        :returns: list of Records
        """

        return self.execute(sql.TOTAL_TABLE_SIZE)

    def total_indexes_size(self):
        """
        Show the total size of all the indexes on each table, descending by
        size.

        Record(
            table='pgbench_accounts',
            index_size='2208 kB'
        )

        :returns: list of Records
        """

        return self.execute(sql.TOTAL_INDEXES_SIZE)

    def table_size(self):
        """
        Show the size of the tables (excluding indexes), descending by size.

        :returns: list
        """

        return self.execute(sql.TABLE_SIZE)

    def index_size(self):
        """
        Show the size of indexes, descending by size.

        :returns: list
        """

        return self.execute(sql.INDEX_SIZE)

    def total_index_size(self):
        """
        Show the total size of all indexes.

        Record(
            size='2240 kB'
        )

        :returns: list of Records
        """

        return self.execute(sql.TOTAL_INDEX_SIZE)

    def locks(self):
        """
        Display queries with active locks.

        Record(
            procpid=31776,
            relname=None,
            transactionid=None,
            granted=True,
            query_snippet='select * from hello;',
            age=datetime.timedelta(0, 0, 288174),
        )

        :returns: list of Records
        """

        return self.execute(
            sql.LOCKS.format(
                pid_column=self.pid_column,
                query_column=self.query_column
            )
        )

    def table_indexes_size(self):
        """
        Show the total size of all the indexes on each table, descending by
        size.

        Record(
            table='pgbench_accounts',
            index_size='2208 kB'
        )

        :returns: list of Records
        """

        return self.execute(sql.TABLE_INDEXES_SIZE)

    def ps(self):
        """
        View active queries with execution time.

        Record(
            pid=28023,
            source='pgbench',
            running_for=datetime.timedelta(0, 0, 288174),
            waiting=0,
            query='UPDATE pgbench_accounts SET abalance = abalance + 423;'
        )

        :returns: list of Records
        """

        if self.is_pg_at_least_nine_two():
            idle = "AND state <> 'idle'"
        else:
            idle = "AND current_query <> '<IDLE>'"

        return self.execute(
            sql.PS.format(
                pid_column=self.pid_column,
                query_column=self.query_column,
                idle=idle
            )
        )

    def version(self):
        """
        Get the Postgres server version.

        Record(
            version='PostgreSQL 9.3.3 on x86_64-apple-darwin13.0.0'
        )

        :returns: list of Records
        """

        return self.execute(sql.VERSION)
